@@ -6,6 +6,7 @@ from typing import Any, Optional, Type
 from bokeh.models import Renderer
 from bokeh.plotting import Figure
 from matplotlib.figure import Figure as matplotlibFigure
+from plotly import graph_objects as go
 
 
 @dataclass
@@ -101,6 +102,16 @@ class PlotGenerator(ABC):
         """
         pass
 
+    @abstractmethod
+    def plot_plotly(self, figure: go.Figure, data, **kwargs) -> None:
+        """
+            plot the given data within the plot.ly Figure
+        :param figure: plot.ly figure
+        :param data: plot data
+        :param kwargs: extra parameters for plotting
+        """
+        pass
+
 
 @dataclass(frozen=True, eq=True)
 class PlotSpec:
