@@ -1,4 +1,4 @@
-import os
+import logging
 from datetime import datetime
 from typing import List, TypeVar, Optional, ContextManager, Iterable, Union, Any
 from typing import Set
@@ -6,7 +6,7 @@ from typing import Set
 import jsonpickle
 import pandas as pd
 from pandas import DataFrame
-from sqlalchemy import create_engine, desc
+from sqlalchemy import desc
 from sqlalchemy.exc import DBAPIError
 from sqlalchemy.orm import sessionmaker, Session
 from sqlalchemy.sql import Selectable
@@ -30,18 +30,19 @@ from entropylab.api.data_writer import (
     PlotSpec,
     NodeData,
 )
+from entropylab.config import settings
 from entropylab.instruments.instrument_driver import Function, Parameter
 from entropylab.instruments.lab_topology import (
     PersistentLabDB,
     DriverType,
     ResourceRecord,
 )
+from entropylab.results_backend.sqlalchemy.db_initializer import _DbInitializer
 from entropylab.results_backend.sqlalchemy.lab_model import (
     Resources,
     ResourcesSnapshots,
 )
 from entropylab.results_backend.sqlalchemy.model import (
-    Base,
     ExperimentTable,
     PlotTable,
     ResultTable,
@@ -50,7 +51,6 @@ from entropylab.results_backend.sqlalchemy.model import (
     NodeTable,
 )
 
-_SQL_ALCHEMY_MEMORY = ":memory:"
 T = TypeVar(
     "T",
 )
@@ -62,24 +62,15 @@ class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
      and DataReader) and lab resources (PersistentLabDB)
     """
 
-    def __init__(self, path=None, echo=False):
+    def __init__(self, path=None, echo=False, **kwargs):
         """
-            Database implementation using SqlAlchemy package for results (DataWriter
-            and DataReader) and lab resources (PersistentLabDB)
-        :param path: database file path (absolute or relative)
+            Initializes database and HDF5 files for an Entropy project
+        :param path: path to directory containing an Entropy project
         :param echo: if True, the database engine will log all statements
         """
         super(SqlAlchemyDB, self).__init__()
-        if path is None:
-            path = _SQL_ALCHEMY_MEMORY
-        else:
-            if path != _SQL_ALCHEMY_MEMORY:
-                dirname = os.path.dirname(path)
-                if dirname and dirname != "":
-                    os.makedirs(dirname, exist_ok=True)
-        path = "sqlite:///" + path
-        self._engine = create_engine(path, echo=echo)
-        Base.metadata.create_all(self._engine)
+        self._enable_hdf5_storage = kwargs.get("enable_hdf5_storage")
+        self._engine, self._storage = _DbInitializer(path, echo=echo).init_db()
         self._Session = sessionmaker(bind=self._engine)
 
     def save_experiment_initial_data(self, initial_data: ExperimentInitialData) -> int:
@@ -99,11 +90,47 @@ class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
                 sess.flush()
 
     def save_result(self, experiment_id: int, result: RawResultData):
+        if result.label is None:
+            raise TypeError("result.label cannot be None")
+        if result.label == "":
+            raise ValueError("result.label cannot be empty")
+        saved_in_hdf5 = False
+        if self.__hdf5_storage_enabled():
+            try:
+                self._storage.save_result(experiment_id, result)
+                saved_in_hdf5 = True
+            except ValueError as ex:
+                raise ValueError(
+                    f"Result already exists (experiment_id='{experiment_id}', "
+                    f"stage='{result.stage}', label='{result.label}'"
+                ) from ex
+            except RuntimeError:
+                logging.exception("Error saving result to HDF5")
+                saved_in_hdf5 = False
         transaction = ResultTable.from_model(experiment_id, result)
+        transaction.saved_in_hdf5 = saved_in_hdf5
         return self._execute_transaction(transaction)
 
     def save_metadata(self, experiment_id: int, metadata: Metadata):
+        if metadata.label is None:
+            raise TypeError("metadata.label cannot be None")
+        if metadata.label == "":
+            raise ValueError("metadata.label cannot be empty")
+        saved_in_hdf5 = False
+        if self.__hdf5_storage_enabled():
+            try:
+                self._storage.save_metadata(experiment_id, metadata)
+                saved_in_hdf5 = True
+            except ValueError as ex:
+                raise ValueError(
+                    f"Metadata already exists (experiment_id='{experiment_id}', "
+                    f"stage='{metadata.stage}', label='{metadata.label}'"
+                ) from ex
+            except RuntimeError:
+                logging.exception("Error saving metadata to HDF5.")
+                saved_in_hdf5 = False
         transaction = MetadataTable.from_model(experiment_id, metadata)
+        transaction.saved_in_hdf5 = saved_in_hdf5
         return self._execute_transaction(transaction)
 
     def save_debug(self, experiment_id: int, debug: Debug):
@@ -120,8 +147,17 @@ class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
 
     def get_experiments_range(self, starting_from_index: int, count: int) -> DataFrame:
         with self._session_maker() as sess:
-            query = sess.query(ExperimentTable).slice(
-                starting_from_index, starting_from_index + count
+            query = (
+                sess.query(ExperimentTable)
+                .with_entities(
+                    ExperimentTable.id,
+                    ExperimentTable.label,
+                    ExperimentTable.start_time,
+                    ExperimentTable.end_time,
+                    ExperimentTable.user,
+                    ExperimentTable.success,
+                )
+                .slice(starting_from_index, starting_from_index + count)
             )
             return self._query_pandas(query)
 
@@ -159,6 +195,20 @@ class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
         experiment_id: Optional[int] = None,
         label: Optional[str] = None,
         stage: Optional[int] = None,
+    ) -> Iterable[ResultRecord]:
+        if self.__hdf5_storage_enabled():
+            return self._storage.get_result_records(experiment_id, stage, label)
+        else:
+            return self.__get_results_from_sqlalchemy(experiment_id, label, stage)
+
+        pass
+
+    def __get_results_from_sqlalchemy(
+        self,
+        experiment_id: Optional[int] = None,
+        label: Optional[str] = None,
+        stage: Optional[int] = None,
+        saved_in_hdf5: Optional[bool] = None,
     ) -> Iterable[ResultRecord]:
         with self._session_maker() as sess:
             query = sess.query(ResultTable)
@@ -168,6 +218,8 @@ class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
                 query = query.filter(ResultTable.label == str(label))
             if stage is not None:
                 query = query.filter(ResultTable.stage == int(stage))
+            if self.__hdf5_storage_enabled() and saved_in_hdf5 is not None:
+                query = query.filter(ResultTable.saved_in_hdf5 == bool(saved_in_hdf5))
             return [item.to_record() for item in query.all()]
 
     def get_metadata_records(
@@ -231,6 +283,14 @@ class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
 
     def get_last_result_of_experiment(
         self, experiment_id: int
+    ) -> Optional[ResultRecord]:
+        if self.__hdf5_storage_enabled():
+            return self._storage.get_last_result_of_experiment(experiment_id)
+        else:
+            return self.__get_last_result_of_experiment_from_sqlalchemy(experiment_id)
+
+    def __get_last_result_of_experiment_from_sqlalchemy(
+        self, experiment_id: int
     ) -> Optional[ResultRecord]:
         with self._session_maker() as sess:
             query = (
@@ -414,3 +474,13 @@ class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
                 return {item.name for item in query}
             else:
                 return set()
+
+    def __hdf5_storage_enabled(self) -> bool:
+        """Feature toggle for 'hdf5 storage' feature
+
+        Class member set in __init__() overrides config setting"""
+        if self._enable_hdf5_storage is None:
+            enabled = settings.get("toggles.hdf5_storage", True)
+        else:
+            enabled = self._enable_hdf5_storage
+        return enabled
