@@ -35,13 +35,14 @@ from entropylab.api.data_writer import (
     PlotSpec,
     NodeData,
 )
+from entropylab.api.errors import EntropyError
 from entropylab.logger import logger
 
 
 def _get_class(module_name, class_name):
     module = importlib.import_module(module_name)
     if not hasattr(module, class_name):
-        raise Exception("class {} is not in {}".format(class_name, module_name))
+        raise EntropyError("class {} is not in {}".format(class_name, module_name))
     logger.debug("reading class {} from module {}".format(class_name, module_name))
     cls = getattr(module, class_name)
     return cls
@@ -120,7 +121,7 @@ class ExperimentTable(Base):
 
 
 class ResultDataType(enum.Enum):
-    """ Numeric values """
+    """Numeric values"""
 
     Pickled = 1
     String = 2
@@ -250,7 +251,7 @@ class PlotTable(Base):
         return f"<PlotSpec(id='{self.id}')>"
 
     def to_record(self) -> PlotRecord:
-        data = _decode_serialized_data(self.data, self.data_type)
+        data = _decode_serialized_data(self.plot_data, self.data_type)
         generator = _get_class(self.generator_module, self.generator_class)
         return PlotRecord(
             experiment_id=self.experiment_id,
